@@ -276,8 +276,7 @@ aggregate2.Matrix<-function(x,groupings=NULL,form=NULL,fun=sum,...)
 #'
 #'Note that \code{NA} values will match other \code{NA} values.
 #'
-#'@param x \code{Matrix} or matrix-like object
-#'@param y \code{Matrix} or matrix-like object
+#'@param x,y \code{Matrix} or matrix-like object
 #'@param by.x vector indicating the names to match from \code{Matrix} x
 #'@param by.y vector indicating the names to match from \code{Matrix} y
 #'@param all.x logical; if \code{TRUE}, then each value in \code{x} will be 
@@ -287,6 +286,9 @@ aggregate2.Matrix<-function(x,groupings=NULL,form=NULL,fun=sum,...)
 #'@param out.class the class of the output object.  Defaults to the class of x. 
 #'  Note that some output classes are not possible due to R coercion
 #'  capabilities, such as converting a character matrix to a Matrix.
+#'@param fill.x,fill.y the value to put in merged columns where there is no match.
+#'  Defaults to 0/FALSE for sparse matrices in order to preserve sparsity, NA for
+#'  all other classes
 #'@param ... arguments to be passed to or from methods.  Currently ignored
 #'@export
 #'@export merge.Matrix
@@ -343,17 +345,20 @@ aggregate2.Matrix<-function(x,groupings=NULL,form=NULL,fun=sum,...)
 #'  by='key',all=TRUE,allow.cartesian=TRUE)})
 #'}
 #'
-merge.Matrix<-function(x,y,by.x,by.y,all.x=TRUE,all.y=TRUE,out.class=class(x),...)
+merge.Matrix<-function(x,y,by.x,by.y,all.x=TRUE,all.y=TRUE,out.class=class(x),
+                       fill.x=ifelse(is(x,'sparseMatrix'),FALSE,NA),fill.y=fill.x,...)
 {
   requireNamespace('grr')
   if(is.null(dim(x)))
     return(grr::matches(by.x,by.y,all.x,all.y,indexes=FALSE))
   indices<-grr::matches(by.x,by.y,all.x,all.y,nomatch = NULL)
-  x<-rBind(x,NA)
-  y<-rBind(y,NA)
+  x<-rBind(x,fill.x)
+  x<-as(grr::extract(x,indices$x),out.class)
+  
+  y<-rBind(y,fill.y)
   if(!is.null(colnames(x)) & !is.null(colnames(y)))
     colnames(y)[colnames(y) %in% colnames(x)]<-paste('y',colnames(y)[colnames(y) %in% colnames(x)],sep='.')
-  x<-as(grr::extract(x,indices$x),out.class)
+
   y<-as(grr::extract(y,indices$y),out.class)
   result<-cbind2(x,y)
   return(result)
@@ -382,6 +387,9 @@ join.Matrix<-merge.Matrix
 #' @param x,... Objects to combine.  If the first argument is a list and 
 #'   \code{..} is unpopulated, the objects in that list will be combined.
 #' @param fill value with which to fill unmatched columns
+#' @param out.class the class of the output object.  Defaults to the class of x. 
+#'  Note that some output classes are not possible due to R coercion
+#'  capabilities, such as converting a character matrix to a Matrix.
 #' @return a single object of the same class as the first input, or of class
 #'   \code{matrix} if the first object is one dimensional
 #' @seealso \code{\link[plyr]{rbind.fill}}
@@ -404,38 +412,33 @@ join.Matrix<-merge.Matrix
 #' colnames(m2)<-3:102
 #' system.time(b<-rBind.fill(m,m2))
 #' 
-rBind.fill<-function(x,...,fill=NA)
+rBind.fill<-function(x,...,fill=NULL,out.class=class(rbind(x,x)))
 {
   if (is.list(x) && !is.data.frame(x) && missing(...)) {
-    Reduce(function (x,y) rBind.fill.internal(x,y,fill),x)
+    Reduce(function (x,y) rBind.fill.internal(x,y,fill,out.class),x)
   }
   else {
-    Reduce(function (x,y) rBind.fill.internal(x,y,fill),list(x,...))
+    Reduce(function (x,y) rBind.fill.internal(x,y,fill,out.class),list(x,...))
   }
 }
 
-rBind.fill.internal<-function(x,y,fill=NA)
+rBind.fill.internal<-function(x,y,fill,out.class)
 {
+  out.class<-force(out.class)
+  fillMissing<-is.null(fill)
+  if(fillMissing)
+    fill<-if(is(x,'Matrix')) 0 else NA
   if (is.null(nrow(x)))
        x<-matrix(x,nrow=1,dimnames=list(NULL,names(x)))
   if (is.null(nrow(y)))
       y<-matrix(y,nrow=1,dimnames=list(NULL,names(y)))
-  y<-
-  {
-    if('data.frame' %in% is(x) && ('Matrix' %in% is(y)))
-      as.data.frame(y)
-    else
-    if('Matrix' %in% is(x))
-      as(y,'Matrix')
-    else
-      y
-  }
-
+  
   nullNames<-FALSE
+  #Cannot currently handle duplicate column names
   if(is.null(colnames(x)))
-    colnames(x)<-colnames(y)[1:ncol(x)]
+    colnames(x)<-make.names(colnames(y)[1:ncol(x)],unique = TRUE)
   if(is.null(colnames(y)))
-     colnames(y)<-colnames(x)[1:ncol(y)]
+     colnames(y)<-make.names(colnames(x)[1:ncol(y)],unique = TRUE)
   if(is.null(colnames(x)))
   {
     nullNames<-TRUE
@@ -443,13 +446,21 @@ rBind.fill.internal<-function(x,y,fill=NA)
     colnames(y)<-1:ncol(y)
   }
   ymiss<-colnames(x)[which(is.na(match(colnames(x),colnames(y))))]
-  ybind<-matrix(fill,nrow=nrow(y),ncol=length(ymiss))
+  ybind<-rsparsematrix(nrow=nrow(y),ncol=length(ymiss),0)
   colnames(ybind)<-ymiss
+  if(!fillMissing)
+    ybind[seq_along(ybind)]<-fill
   xmiss<-colnames(y)[which(is.na(match(colnames(y),colnames(x))))]
-  xbind<-matrix(fill,nrow=nrow(x),ncol=length(xmiss))
+  xbind<-rsparsematrix(nrow=nrow(x),ncol=length(xmiss),0)
   colnames(xbind)<-xmiss
-  x<-cbind2(x,xbind)
-  y<-cbind2(y,ybind)
+  if(!fillMissing)
+    xbind[seq_along(xbind)]<-fill
+  if (ncol(xbind>0))
+    x<-cbind2(x,xbind)
+  if(ncol(ybind)>0)
+    y<-cbind2(y,ybind)
+  y<-as(y,out.class)
+  x<-as(x,out.class)
   result<-rbind2(x,y[,order(match(colnames(y),colnames(x)))])
   if(nullNames)
     colnames(result)<-NULL
@@ -462,6 +473,8 @@ len<-function (data)
   return(result)
 }
 
+setAs('Matrix','data.frame',function (from) as.data.frame(as.matrix(from)),)
+
 setAs('Matrix','data.frame',function (from) as.data.frame(as.matrix(from)))
 
 setAs('data.frame','dgeMatrix', function (from) as(as.matrix(from),'dgeMatrix'))
@@ -472,3 +485,16 @@ setAs('matrix','data.frame',function (from) as.data.frame(from))
 
 setAs('vector','data.frame',function (from) data.frame(from))
 
+setMethod("cbind2",c('data.frame','Matrix'),function (x,y) 
+{
+  y<-as.matrix(y)
+  cbind2(x,y)
+}
+)
+
+setMethod("cbind2",c('Matrix','data.frame'),function (x,y) 
+{
+  y<-as.matrix(y)
+  cbind2(x,y)
+}
+)
